import sys
import cv2
import numpy as np
import subprocess
import itertools
import uuid  # Added for unique identifiers
import os  # Added for file deletion
from concurrent.futures import ThreadPoolExecutor
import time

def convert_to_jpg(input_filename, output_filename):
    # Convert the first page of the input image to jpg
    command = ["magick", "convert", f"{input_filename}[0]", f"{output_filename}"]
    subprocess.run(command, check=True)

def align_channels(red, green, blue, uniq):
    # Convert to jpg format (use only the first page)
    red_jpg = f"red_{uniq}_{red}.jpg"
    green_jpg = f"green_{uniq}_{green}.jpg"
    blue_jpg = f"blue_{uniq}_{blue}.jpg"

    convert_to_jpg(red, red_jpg)
    convert_to_jpg(green, green_jpg)
    convert_to_jpg(blue, blue_jpg)

    # Load images
    img_red = cv2.imread(red_jpg, cv2.IMREAD_GRAYSCALE)
    img_green = cv2.imread(green_jpg, cv2.IMREAD_GRAYSCALE)
    img_blue = cv2.imread(blue_jpg, cv2.IMREAD_GRAYSCALE)

    # Create SIFT detector
    sift = cv2.SIFT_create()

    # Find keypoints and descriptors with SIFT
    kp_red, des_red = sift.detectAndCompute(img_red, None)
    kp_green, des_green = sift.detectAndCompute(img_green, None)
    kp_blue, des_blue = sift.detectAndCompute(img_blue, None)

    # FLANN parameters
    FLANN_INDEX_KDTREE = 1
    index_params = dict(algorithm=FLANN_INDEX_KDTREE, trees=5)
    search_params = dict(checks=50)  # or pass empty dictionary

    # FLANN matcher
    flann = cv2.FlannBasedMatcher(index_params, search_params)

    # Match descriptors using KNN
    matches_green = flann.knnMatch(des_green, des_red, k=2)
    matches_blue = flann.knnMatch(des_blue, des_red, k=2)

    # Need to draw only good matches, so create a mask
    matches_mask_green = [[0, 0] for _ in range(len(matches_green))]
    matches_mask_blue = [[0, 0] for _ in range(len(matches_blue))]

    # Ratio test as per Lowe's paper
    for i, (m, n) in enumerate(matches_green):
        if m.distance < 0.7 * n.distance:
            matches_mask_green[i] = [1, 0]

    for i, (m, n) in enumerate(matches_blue):
        if m.distance < 0.7 * n.distance:
            matches_mask_blue[i] = [1, 0]

    # Apply transformation to green and blue channels
    src_pts_green = np.float32([kp_green[m[0].queryIdx].pt for m, mask in zip(matches_green, matches_mask_green) if mask[0]])
    dst_pts_green = np.float32([kp_red[m[0].trainIdx].pt for m, mask in zip(matches_green, matches_mask_green) if mask[0]])

    src_pts_blue = np.float32([kp_blue[m[0].queryIdx].pt for m, mask in zip(matches_blue, matches_mask_blue) if mask[0]])
    dst_pts_blue = np.float32([kp_red[m[0].trainIdx].pt for m, mask in zip(matches_blue, matches_mask_blue) if mask[0]])


    warp_matrix_green, _ = cv2.findHomography(src_pts_green, dst_pts_green, cv2.RANSAC, 5.0)
    warp_matrix_blue, _ = cv2.findHomography(src_pts_blue, dst_pts_blue, cv2.RANSAC, 5.0)

    img_green_aligned = cv2.warpPerspective(img_green, warp_matrix_green, img_green.shape[::-1], flags=cv2.INTER_LINEAR)
    img_blue_aligned = cv2.warpPerspective(img_blue, warp_matrix_blue, img_blue.shape[::-1], flags=cv2.INTER_LINEAR)

    # Save aligned images
    aligned_green_filename = f"aligned_green_{uniq}_{green}.jpg"
    aligned_blue_filename = f"aligned_blue_{uniq}_{blue}.jpg"
    cv2.imwrite(aligned_green_filename, img_green_aligned)
    cv2.imwrite(aligned_blue_filename, img_blue_aligned)

    # Delete intermediate files
    print(f"Trying to remove {red_jpg}")
    os.remove(red_jpg)
    os.remove(green_jpg)
    os.remove(blue_jpg)

    return aligned_green_filename, aligned_blue_filename

def combine_and_save(combo):
    red, green, blue = combo
    h = hash(f"Doing {red} {green} {blue}")
    print(f"Doing {red} {green} {blue}")
    aligned_green, aligned_blue = align_channels(red, green, blue, h)
    time.sleep(1)
    output_filename = f"combined_{red}_{green}_{blue}.jpg"
    command = [
        "magick", "convert",
        f"{red}[0]", f"{aligned_green}", f"{aligned_blue}",
        "-auto-level", "-combine", "-set", "colorspace", "sRGB",
        f"{output_filename}"
    ]

    try:
        subprocess.run(command, check=True)
        print(f"Images {red}, {aligned_green}, {aligned_blue} combined successfully. Output: {output_filename}")
    except subprocess.CalledProcessError as e:
        print(f"Error combining images: {e}")

    # Delete intermediate files
    print(f"Trying to remove {aligned_green}")
    os.remove(aligned_green)
    os.remove(aligned_blue)

def combine_channels_parallel(image_list):
    combinations = list(itertools.permutations(image_list, 3))

    with ThreadPoolExecutor(max_workers=6) as executor:
        executor.map(combine_and_save, combinations)

if __name__ == "__main__":
    # Check if at least 3 image filenames are provided
    if len(sys.argv) < 4:
        print("Please provide at least 3 image filenames.")
        sys.exit(1)

    # Get image filenames from command-line arguments
    image_filenames = sys.argv[1:]

    combine_channels_parallel(image_filenames)

